{****************************************************************************
    Delphi component
    Memory-mapped file
    Version 1.01 Beta
    Copyright (c) by Togushev Dmitry 1999 y.
       e-mail: j190376@hotmail.com
****************************************************************************}

unit JMMFile;

interface

uses
   Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

const
   errFunctionNotEmplemented = -100;
   errFileAlwaysOpened = -101;
   errFileTooBig       = -102;
   errOutOfSize        = -103;
   errFileReadOnly     = -104;
   errFileNotOpened    = -105;
   errInvalidDirection = -106;
   errZeroLength       = -107;

   dmWrite             = 1;
   dmRead              = 0;

type
   EMemMap = class(Exception);

   TAccessMode = (amRead, amReadWrite);
   TShareMode  = (smNoShare, smReadShare, smWriteShare, smAllShare);
   TOpenMode   = (cmCreateNew, cmCreateAlways, cmOpenExisting, cmOpenAlways, cmTruncExist);

   TJMemMapFile = class(TComponent)
   protected
      FFileName : string;
      FActive   : boolean;
      FFirstIndex : DWORD;
      FAccessMode : TAccessMode;
      FShareMode  : TShareMode;
      FOpenMode   : TOpenMode;
      FBuffering  : boolean;

      FFileHandle : THandle;
      FMapHandle  : THandle;
      FMemPointer : pointer;
      FViewOffset : DWORD;
      FViewSize   : DWORD;
      FFileSize   : DWORD;
      FAddSize    : DWORD;
      FMapAccess  : DWORD;

      FLastError  : integer;

      procedure Open;                                                        virtual;
      procedure Close;                                                       virtual;

      function  MovePoint(Offset : DWORD; UnitSize : DWORD) : pointer;       virtual;
      function  MovePointByIndex(Index : DWORD; UnitSize : DWORD) : pointer; virtual;
      procedure OpenView ( ReqOffset : DWORD);                               virtual;
      procedure SetFileName( AFileName : string);                            virtual;
      procedure SetActive( Value : boolean);                                 virtual;
      procedure SetFirstIndex( Value : DWORD);                               virtual;
      procedure SetAccessMode( Value : TAccessMode);                         virtual;
      procedure SetShareMode( Value : TShareMode);                           virtual;
      procedure SetOpenMode( Value : TOpenMode);                             virtual;
      procedure SetBuffering( Value : boolean);                              virtual;

      procedure SetByte(I : DWORD; Value : byte);                            virtual;
      function  GetByte(I : DWORD) : byte;                                   virtual;
      procedure SetWord(I : DWORD; Value : word);                            virtual;
      function  GetWord(I : DWORD) : word;                                   virtual;
      procedure SetLong(I : DWORD; Value : longint);                         virtual;
      function  GetLong(I : DWORD) : longint;                                virtual;
      procedure SetAChar(I : DWORD; Value : AnsiChar);                       virtual;
      function  GetAChar(I : DWORD) : AnsiChar;                              virtual;
      procedure SetWChar(I : DWORD; Value : WideChar);                       virtual;
      function  GetWChar(I : DWORD) : WideChar;                              virtual;
      procedure ArrayMove(Offset, Size : DWORD; Buffer : pointer; Direct : integer); virtual;

      procedure SetError(AError : integer);                                  virtual;

      property  FError : integer read FLastError write SetError;
   public
      constructor Create( AOwner : TComponent);                              override;
      destructor  Destroy;                                                   override;

      procedure ReadArray(Offset, Size : DWORD; Buffer : pointer);          virtual;
      procedure WriteArray(Offset, Size : DWORD; Buffer : pointer);         virtual;
      procedure FlushFileBuffers;                                           virtual;
      procedure SetFileLength(Length : DWORD);                              virtual;

      property  AsByte [I : DWORD] : byte read GetByte write SetByte;
      property  AsWord [I : DWord] : word read GetWord write SetWord;
      property  AsLong [I : DWord] : longint read GetLong write SetLong;     default;
      property  AsChar [I : DWord] : AnsiChar read GetAChar write SetAChar;
      property  AsWChar[I : DWord] : WideChar read GetWChar write SetWChar;

      property  LastError : integer read FLastError;
   published
      property  FileName   : string read FFileName write SetFileName;
      property  Active     : boolean read FActive write SetActive;
      property  FileSize   : DWORD read FFileSize;
      property  AddSize    : DWORD       read FAddSize    write FAddSize      default 0;
      property  FirstIndex : DWORD read FFirstIndex       write SetFirstIndex default 1;
      property  AccessMode : TAccessMode read FAccessMode write SetAccessMode default amRead;
      property  ShareMode  : TShareMode  read FShareMode  write SetShareMode  default smReadShare;
      property  OpenMode   : TOpenMode   read FOpenMode   write SetOpenMode   default cmOpenExisting;
      property  Buffering  : boolean     read FBuffering  write SetBuffering  default true;
   end;

procedure Register;

implementation

const
   ResursGranularity : DWORD = 0;

procedure Register;
begin
  RegisterComponents('JComp', [TJMemMapFile]);
end;

{********************************************************************************}

constructor TJMemMapFile.Create;
var
   SysInf : _SYSTEM_INFO;

begin
   inherited;
   FFileName := '';
   FActive   := false;

   FAccessMode := amRead;
   FShareMode  := smReadShare;
   FOpenMode   := cmOpenExisting;
   FBuffering  := true;

   FFileHandle := 0;
   FMapHandle  := 0;
   FMemPointer := nil;
   FFileSize   := 0;
   AddSize     := 0;

   FLastError  := 0;

   FFirstIndex := 1;

   GetSystemInfo(SysInf);
   ResursGranularity := SysInf.dwAllocationGranularity;
end;

destructor  TJMemMapFile.Destroy;
begin
   try
      Close;
   except
   end;
   inherited;
end;

{********************************************************************************}

function TJMemMapFile.MovePoint(Offset : DWORD; UnitSize : DWORD) : pointer;
begin
   if not FActive then
      FError := errFileNotOpened;

   if ((Offset+UnitSize) > FFileSize)or(Offset > FFileSize) then
      FError := errOutOfSize;

   OpenView(Offset);

   result := pointer(DWORD(FMemPointer)+(Offset-FViewOffset));
end;

function  TJMemMapFile.MovePointByIndex(Index : DWORD; UnitSize : DWORD) : pointer;
begin
   Index := Index - FFirstIndex;
   result := MovePoint(Index*UnitSize, UnitSize);
end;

procedure TJMemMapFile.OpenView ( ReqOffset : DWORD);
begin
   SetLastError(0);
   if ((FViewOffset + FViewSize) > ReqOffset)and(FViewOffset <= ReqOffset) and (FMemPointer <> nil) then
      exit;

   if FMemPointer <> nil then begin
      UnmapViewOfFile(FMemPointer);
      FMemPointer := nil;
      FError := GetLastError;
   end;

   FViewOffset := ResursGranularity * (ReqOffset div ResursGranularity);
   if FFileSize - FViewOffset < ResursGranularity then
      FViewSize := FFileSize - FViewOffset
   else
      FViewSize := ResursGranularity;

   FMemPointer := MapViewOfFile(FMapHandle, FMapAccess, 0, FViewOffset, FViewSize);
   if FMemPointer = nil then
      FError := GetLastError;
end;

procedure TJMemMapFile.SetFileName( AFileName : string);
begin
   if FActive then
      FError := errFileAlwaysOpened;
   FFileName := AFileName;
end;

procedure TJMemMapFile.SetActive( Value : boolean);
begin
   if FActive <> Value then
      if Value then 
         Open
      else
         Close;
end;

procedure TJMemMapFile.SetFirstIndex( Value : DWORD);
begin
   if FActive then
      FError := errFileAlwaysOpened;
   FFirstIndex := Value;
end;

procedure TJMemMapFile.SetAccessMode( Value : TAccessMode);
begin
   if FActive then
      FError := errFileAlwaysOpened;
   FAccessMode := Value;
end;

procedure TJMemMapFile.SetShareMode( Value : TShareMode);
begin
   if FActive then
      FError := errFileAlwaysOpened;
   FShareMode := Value;
end;

procedure TJMemMapFile.SetOpenMode( Value : TOpenMode);
begin
   if FActive then
      FError := errFileAlwaysOpened;
   if Value = cmTruncExist then
      if FAccessMode<>amReadWrite then
         exit;
   FOpenMode := Value;
end;

procedure TJMemMapFile.SetBuffering( Value : boolean);
begin
   if FActive then
      FError := errFileAlwaysOpened;
   FBuffering := Value;
end;

{*******************************************************************************}

function  TJMemMapFile.GetByte(I : DWORD) : byte;
var
   Unt : ^byte;

begin
   if not FActive then
      FError := errFileNotOpened;
   Unt := MovePointByIndex(I, sizeof(byte));
   result := Unt^;
end;

function  TJMemMapFile.GetWord(I : DWORD) : word;
var
   Unt : ^word;

begin
   if not FActive then
      FError := errFileNotOpened;
   Unt := MovePointByIndex(I, sizeof(word));
   result := Unt^;
end;

function  TJMemMapFile.GetLong(I : DWORD) : longint;
var
   Unt : ^longint;

begin
   if not FActive then
      FError := errFileNotOpened;
   Unt := MovePointByIndex(I, sizeof(longint));
   result := Unt^;
end;

function  TJMemMapFile.GetAChar(I : DWORD) : AnsiChar;
var
   Unt : ^AnsiChar;

begin
   if not FActive then
      FError := errFileNotOpened;
   Unt := MovePointByIndex(I, sizeof(AnsiChar));
   result := Unt^;
end;

function  TJMemMapFile.GetWChar(I : DWORD) : WideChar;
var
   Unt : ^WideChar;

begin
   if not FActive then
      FError := errFileNotOpened;
   Unt := MovePointByIndex(I, sizeof(WideChar));
   result := Unt^;
end;

procedure TJMemMapFile.SetByte(I : DWORD; Value : byte);
var
   Unt : ^byte;

begin
   if not FActive then
      FError := errFileNotOpened;
   if FAccessMode <> amReadWrite then
      FError := errFileReadOnly;
   Unt := MovePointByIndex(I, sizeof(byte));
   Unt^ := Value;
end;

procedure TJMemMapFile.SetWord(I : DWORD; Value : word);
var
   Unt : ^word;

begin
   if not FActive then
      FError := errFileNotOpened;
   if FAccessMode <> amReadWrite then
      FError := errFileReadOnly;
   Unt := MovePointByIndex(I, sizeof(word));
   Unt^ := Value;
end;

procedure TJMemMapFile.SetLong(I : DWORD; Value : longint);
var
   Unt : ^longint;

begin
   if not FActive then
      FError := errFileNotOpened;
   if FAccessMode <> amReadWrite then
      FError := errFileReadOnly;
   Unt := MovePointByIndex(I, sizeof(longint));
   Unt^ := Value;
end;

procedure TJMemMapFile.SetAChar(I : DWORD; Value : AnsiChar);
var
   Unt : ^AnsiChar;

begin
   if not FActive then
      FError := errFileNotOpened;
   if FAccessMode <> amReadWrite then
      FError := errFileReadOnly;
   Unt := MovePointByIndex(I, sizeof(AnsiChar));
   Unt^ := Value;
end;

procedure TJMemMapFile.SetWChar(I : DWORD; Value : WideChar);
var
   Unt : ^WideChar;

begin
   if not FActive then
      FError := errFileNotOpened;
   if FAccessMode <> amReadWrite then
      FError := errFileReadOnly;
   Unt := MovePointByIndex(I, sizeof(WideChar));
   Unt^ := Value;
end;

procedure TJMemMapFile.ArrayMove(Offset, Size : DWORD; Buffer : pointer; Direct : integer);
var
   Unt : pointer;
   CopySize : DWORD;

begin
   if not FActive then
      FError := errFileNotOpened;

   if (Offset + Size) > FFileSize then
      FError := errOutOfSize;

   Unt := MovePoint(Offset, Size);
   while Size <> 0 do begin
      if Size > FViewSize then
         CopySize := FViewSize
      else
         CopySize := Size;

      Size := Size - CopySize;
      Offset := Offset + CopySize;

      case Direct of
        dmWrite : move(Buffer^, Unt^, CopySize);
        dmRead  : move(Unt^, Buffer^, CopySize);
        else
           FError := errInvalidDirection;
      end;

      Buffer := pointer(DWORD(Buffer)+CopySize);
      Unt := MovePoint(Offset, Size);
   end;
end;

{*******************************************************************************}

procedure TJMemMapFile.SetError(AError : integer);
begin
   FLastError := AError;
   if AError <> ERROR_SUCCESS then
      raise EMemMap.Create('Error number '+inttostr(AError)+' in memeory map file');
end;

{********************************************************************************}

procedure TJMemMapFile.Open;
var
   fn : array [0..1024] of char;
   HDM : DWORD;
   FileAccess : DWORD;
   PageAccess : DWORD;
   FileShare  : DWORD;
   FileOpMpde : DWORD;
   FS         : DWORD;
   FileAttribute : DWORD;

begin
   SetLastError(0);

   FileAccess := GENERIC_READ;
   PageAccess := PAGE_READONLY;
   FMapAccess  := FILE_MAP_READ;
   FS         := 0;
   case FAccessMode of
     amReadWrite : begin
        FileAccess := GENERIC_READ or GENERIC_WRITE;
        PageAccess := PAGE_READWRITE;
        FMapAccess  := FILE_MAP_WRITE;
     end;
     amRead      : begin
        FileAccess := GENERIC_READ;
        PageAccess := PAGE_READONLY;
        FMapAccess  := FILE_MAP_READ;
        FS         := 0;
     end;
   end;

   FileShare := 0;
   case FShareMode of
     smWriteShare : FileShare := FILE_SHARE_WRITE;
     smNoShare    : FileShare := 0;
     smReadShare  : FileShare := FILE_SHARE_READ;
     smAllShare   : FileShare := FILE_SHARE_WRITE or FILE_SHARE_READ;
   end;

   FileOpMpde := OPEN_EXISTING;
   case FOpenMode of
     cmCreateAlways : FileOpMpde := CREATE_ALWAYS;
     cmOpenExisting : FileOpMpde := OPEN_EXISTING;
     cmCreateNew    : FileOpMpde := CREATE_NEW;
     cmOpenAlways   : FileOpMpde := OPEN_ALWAYS;
     cmTruncExist   : FileOpMpde := TRUNCATE_EXISTING;
   end;

   FileAttribute := FILE_ATTRIBUTE_NORMAL;
   if not FBuffering then
      FileAttribute := FileAttribute or FILE_FLAG_NO_BUFFERING;

   strplcopy(fn, FFileName, sizeof(fn));
   FFileHandle := CreateFile(fn, FileAccess, FileShare, nil, FileOpMpde, FileAttribute, 0);
   if FFileHandle = INVALID_HANDLE_VALUE then
      FError := GetLastError;

   FActive := true;

   FFileSize := GetFileSize(FFileHandle, @HDM);
   if HDM <> 0 then
      try
         FError := errFileTooBig;
      finally
         Close;
      end
   else
      if FAccessMode = amReadWrite then begin
         FS := FFileSize + FAddSize;
         FFileSize := FS;
      end;

   if (fs=0) and (FFileSize = 0) then
      FError := errZeroLength;

   FMapHandle := CreateFileMapping(FFileHandle, nil, PageAccess, 0, FS, nil);
   if FMapHandle = 0 then
      try
         FError := GetLastError;
      finally
         Close;
      end;

   OpenView(0);
end;

procedure TJMemMapFile.Close;
begin
   SetLastError(0);
   if FActive then begin
      FActive := false;

      if FMemPointer <> nil then begin
         UnmapViewOfFile(FMemPointer);
         FMemPointer := nil;
         FError := GetLastError;
      end;

      if FMapHandle <> 0 then begin
         CloseHandle(FMapHandle);
         FMapHandle := 0;
         FError := GetLastError;
      end;

      if FFileHandle <> 0 then begin
         CloseHandle(FFileHandle);
         FFileHandle := 0;
         FError := GetLastError;
      end;
   end;
end;

{*******************************************************************************}

procedure TJMemMapFile.ReadArray(Offset, Size : DWORD; Buffer : pointer);
begin
   ArrayMove(Offset, Size, Buffer, dmRead);
end;

procedure TJMemMapFile.WriteArray(Offset, Size : DWORD; Buffer : pointer);
begin
   ArrayMove(Offset, Size, Buffer, dmWrite);
end;

procedure TJMemMapFile.FlushFileBuffers;
begin
   SetLastError(0);
   if not FActive then
      FError := errFileNotOpened;

   windows.FlushFileBuffers(FFileHandle);

   FError := GetLastError;
end;

procedure TJMemMapFile.SetFileLength(Length : DWORD);
var
   HDM, FS : DWORD;
   fn : array [0..1024] of char;

begin
   SetLastError(0);
   if FActive then
      FError := errFileAlwaysOpened;

   strplcopy(fn, FFileName, sizeof(fn));
   FFileHandle := CreateFile(fn, GENERIC_READ or GENERIC_WRITE, 0, nil, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
   if FFileHandle = INVALID_HANDLE_VALUE then
      FError := GetLastError;

   HDM := 0;
   FS := GetFileSize(FFileHandle, @HDM);
   if HDM <> 0 then
      try
         FError := errFileTooBig;
      finally
         CloseHandle(FFileHandle);
      end;

   if Length > FS then
      FError := errOutOfSize;

   SetFilePointer(FFileHandle, Length, @HDM, FILE_BEGIN);
   SetEndOfFile(FFileHandle);
   CloseHandle(FFileHandle);
end;

end.

